"""
All constants for the package.
"""

DATE_FORMAT = "%Y-%m-%d"
LOGGER_NAME = "sweetpay-sdk"
OK_STATUS = "OK"

# Define some test data for SE
TEST_CREDIT_SSN = "19500101-0001"
TEST_NOCREDIT_SSN = "19500101-0007"
